"""
This is just a little low level library for fetching keyboard input.
This is a porting of the Numworks module, and add other methods created by others OS (like Omega or Upsilon).
"""

try: from .util.ion import Ion as __Ion, OS_MODE
except ImportError as e:
  if "relative import" not in e.msg:
    raise
  from util.ion import Ion as __Ion, OS_MODE

__name__ = "ion"
__version__ = "2.0"
try:
  with open("README.md", encoding="utf-8") as f: __doc__ = f.read()
  del f
except (FileNotFoundError, OSError): __doc__ = "<unknown>"
__all__ = [
  "keydown",
  "get_keys",
  "battery",
  "battery_level",
  "battery_ischarging",
  "set_brightness",
  "get_brightness",
  "file", # idk what is this
]


### All keys of Numworks
try: from .util.stuff.keys import *
except ImportError as e:
  if "relative import" not in e.msg:
    raise
  from util.stuff.keys import *
# Add all keys in __all__
__all__.extend([i for i in dir() if i.startswith("KEY_")])


### Methods
def keydown(k, /):
  """Return True if the k key is pressed (not release)"""
  key, err = __Ion.call(__Ion.keydown, k)
  if err != None:
    raise err
  return key

def get_keys():
  """Get name of pressed keys"""
  keys, err = __Ion.call(__Ion.get_keys)
  if err != None:
    raise err
  return keys

# All the following functions only give a fake result to give a real look of library
def battery():
  """Return battery voltage"""
  voltage, err = __Ion.call(__Ion.battery)
  if err != None:
    raise err
  return voltage

def battery_level():
  """Return battery level"""
  level, err = __Ion.call(__Ion.battery_level)
  if err != None:
    raise err
  return level

def battery_ischarging():
  """Return True if the battery is charging"""
  charging, err = __Ion.call(__Ion.battery_ischarging)
  if err != None:
    raise err
  return charging

def set_brightness(level, /):
  """Set brightness level of screen"""
  _, err = __Ion.call(__Ion.set_brightness, level)
  if err != None:
    raise err

def get_brightness():
  """Get brightness level of screen"""
  brightness, err = __Ion.call(__Ion.get_brightness)
  if err != None:
    raise err
  return brightness

# I don't know why this exist, but is in source code of Omega and Upsilon
class file:
  SEEK_SET = 0
  SEEK_CUR = 1
  SEEK_END = 2
  def __init__(self, *_, **__):
    raise \
      TypeError(f"cannot create '{self.__class__.__name__}' instances")


### Cleanup
if OS_MODE:
  if OS_MODE < 3:
    del get_keys, battery, battery_level, battery_ischarging, set_brightness, get_brightness
    __all__.remove("get_keys")
    __all__.remove("battery")
    __all__.remove("battery_level")
    __all__.remove("battery_ischarging")
    __all__.remove("set_brightness")
    __all__.remove("get_brightness")

  if OS_MODE == 1:
    del file
    __all__.remove("file")
del OS_MODE



from pynput.keyboard import Listener

from .pynput_patcher import *
from .keys import ALL_KEYS, NUMBER_OF_KEYS
from .focus_checker import FocusChecker
from .common import print_debug


class KeyLogger:
  _listener: Listener = None
  _check_focus = None
  _focused = False
  _keyboard_state = {}
  _error = None

  @staticmethod
  def __init__():
    """
    Start the KeyLogger.
    This is an singleton, so the constructor cannot be called again until after calling .stop().
    """

    if KeyLogger.is_running(): raise RuntimeError("KeyLogger already running")

    def on_press(key):
      if key is None: return # because the key can be None in some cases

      print_debug("Pressed", key)
      if hasattr(key, "char"): key = key.char
      try: KeyLogger._focused = KeyLogger._check_focus()
      except BaseException as e:
        # an error occurs while checking focus, so pass this error to the main thread
        KeyLogger.stop()
        KeyLogger._error = e
        return False

      if KeyLogger._focused:
        for i in range(NUMBER_OF_KEYS):
          k = ALL_KEYS[i]["key"]
          if (k == key or (type(k) in (list, tuple) and any([i == key for i in k]))):
            KeyLogger.set_key(ALL_KEYS[i]["code"], True)

    def on_release(key):
      if key is None: return # because the key can be None is some cases

      print_debug("Released", key)
      if hasattr(key, "char"): key = key.char

      for i in range(NUMBER_OF_KEYS):
        k = ALL_KEYS[i]["key"]
        if (k == key or (type(k) in (list, tuple) and any([i == key for i in k]))):
          KeyLogger.set_key(ALL_KEYS[i]["code"], False)

    KeyLogger._error = None # remove errors
    KeyLogger._keyboard_state = {k["code"]: False for k in ALL_KEYS}
    KeyLogger._check_focus = FocusChecker()
    KeyLogger._listener = Listener(on_press=on_press, on_release=on_release)
    KeyLogger._listener.start()

  @staticmethod
  def raise_if_error():
    """Raise the last error"""

    if KeyLogger._error:
      error = KeyLogger._error
      KeyLogger._error = None # remove the last error after raised it
      raise error

  @staticmethod
  def check_ok(code):
    """Check the Keylogger, the keycode and the focus"""

    if not KeyLogger.is_running(): raise RuntimeError("KeyLogger not running")
    elif type(code) != int: raise TypeError(f"keycode must be an integer, not {type(code).__name__}")
    elif code not in KeyLogger._keyboard_state: raise IndexError(f"key with code '{code}' not found")
    try: KeyLogger._check_focus(True)
    except:
      KeyLogger.stop()
      raise

  @staticmethod
  def stop():
    """Stop the KeyLogger"""

    if not KeyLogger.is_running(): return
    if KeyLogger._listener: KeyLogger._listener.stop()
    KeyLogger._listener = None
    KeyLogger._check_focus = None
    KeyLogger._focused = False
    KeyLogger._keyboard_state = {}

  @staticmethod
  def is_running():
    """Return whether the KeyLogger is running"""
    
    return KeyLogger._listener and KeyLogger._listener.is_alive()

  @staticmethod
  def get_key(code):
    """Get state of a key (is pressed or not) with his keycode"""

    KeyLogger.check_ok(code)
    return KeyLogger._focused and KeyLogger._keyboard_state[code]

  @staticmethod
  def set_key(code, is_pressed, add=False):
    """Set state of a key with his keycode"""

    KeyLogger.check_ok(code)
    KeyLogger._keyboard_state[code] = bool(is_pressed)



from pynput.keyboard import Key

KEY_LEFT: int =             {'code': 0,  'name': 'left',      'key': Key.left}
KEY_UP: int =               {'code': 1,  'name': 'up',        'key': Key.up}
KEY_DOWN: int =             {'code': 2,  'name': 'down',      'key': Key.down}
KEY_RIGHT: int =            {'code': 3,  'name': 'right',     'key': Key.right}
KEY_OK: int =               {'code': 4,  'name': 'OK',        'key': Key.enter}
KEY_BACK: int =             {'code': 5,  'name': 'back',      'key': Key.delete}
KEY_HOME: int =             {'code': 6,  'name': 'home',      'key': Key.esc}
KEY_ONOFF: int =            {'code': 7,  'name': 'onOff',     'key': Key.end}
KEY_SHIFT: int =            {'code': 12, 'name': 'shift',     'key': (Key.shift_l, Key.shift_r)}
KEY_ALPHA: int =            {'code': 13, 'name': 'alpha',     'key': (Key.ctrl_l, Key.ctrl_r)}
KEY_XNT: int =              {'code': 14, 'name': 'xnt',       'key': 'x'}
KEY_VAR: int =              {'code': 15, 'name': 'var',       'key': 'v'}
KEY_TOOLBOX: int =          {'code': 16, 'name': 'toolbox',   'key': '"'}
KEY_BACKSPACE: int =        {'code': 17, 'name': 'backspace', 'key': Key.backspace}
KEY_EXP: int =              {'code': 18, 'name': 'exp',       'key': 'e'}
KEY_LN: int =               {'code': 19, 'name': 'ln',        'key': 'n'}
KEY_LOG: int =              {'code': 20, 'name': 'log',       'key': 'l'}
KEY_IMAGINARY: int =        {'code': 21, 'name': 'imaginary', 'key': 'i'}
KEY_COMMA: int =            {'code': 22, 'name': 'comma',     'key': ','}
KEY_POWER: int =            {'code': 23, 'name': 'power',     'key': '^'}
KEY_SINE: int =             {'code': 24, 'name': 'sin',       'key': 's'}
KEY_COSINE: int =           {'code': 25, 'name': 'cos',       'key': 'c'}
KEY_TANGENT: int =          {'code': 26, 'name': 'tan',       'key': 't'}
KEY_PI: int =               {'code': 27, 'name': 'pi',        'key': 'p'}
KEY_SQRT: int =             {'code': 28, 'name': 'sqrt',      'key': 'r'}
KEY_SQUARE: int =           {'code': 29, 'name': 'square',    'key': '>'}
KEY_SEVEN: int =            {'code': 30, 'name': '7',         'key': '7'}
KEY_EIGHT: int =            {'code': 31, 'name': '8',         'key': '8'}
KEY_NINE: int =             {'code': 32, 'name': '9',         'key': '9'}
KEY_LEFTPARENTHESIS: int =  {'code': 33, 'name': '(',         'key': '('}
KEY_RIGHTPARENTHESIS: int = {'code': 34, 'name': ')',         'key': ')'}
KEY_FOUR: int =             {'code': 36, 'name': '4',         'key': '4'}
KEY_FIVE: int =             {'code': 37, 'name': '5',         'key': '5'}
KEY_SIX: int =              {'code': 38, 'name': '6',         'key': '6'}
KEY_MULTIPLICATION: int =   {'code': 39, 'name': '*',         'key': '*'}
KEY_DIVISION: int =         {'code': 40, 'name': '/',         'key': '/'}
KEY_ONE: int =              {'code': 42, 'name': '1',         'key': '1'}
KEY_TWO: int =              {'code': 43, 'name': '2',         'key': '2'}
KEY_THREE: int =            {'code': 44, 'name': '3',         'key': '3'}
KEY_PLUS: int =             {'code': 45, 'name': '+',         'key': '+'}
KEY_MINUS: int =            {'code': 46, 'name': '-',         'key': '-'}
KEY_ZERO: int =             {'code': 48, 'name': '0',         'key': '0'}
KEY_DOT: int =              {'code': 49, 'name': '.',         'key': '.'}
KEY_EE: int =               {'code': 50, 'name': 'EE',        'key': '!'}
KEY_ANS: int =              {'code': 51, 'name': 'Ans',       'key': 'a'}
KEY_EXE: int =              {'code': 52, 'name': 'EXE',       'key': Key.insert}# TODO: not on macos, find another key

# Put all keys in ALL_KEYS and redefine each key only by its code
ALL_KEYS = []

for n, k in locals().copy().items():
  # Avoid to re-replace variable if file is imported at multiple times
  if n.startswith("KEY_") and type(k) == dict:
    k.update({'field': n})
    locals()[n] = k['code']
    ALL_KEYS.append(k)

NUMBER_OF_KEYS = len(ALL_KEYS)
__all__ = [k['field'] for k in ALL_KEYS]



from ..common import prettywarn, print_debug
import sys, os

# By default it just read kandinsky window (only if is focused)
DISABLE_KANDINSKY_INPUT_ONLY = 'ION_DISABLE_KANDINSKY_INPUT_ONLY' in os.environ
# Option to get input everywhere on system
GET_INPUT_EVERYWHERE = 'ION_ENABLE_GET_INPUT_EVERYWHERE' in os.environ


class BaseFocusChecker:
  """
  Base class for FocusChecker

  following methods must be overrided:
    - check_window(wid, pid, classname, not_classname, contains_title)
    - search_window(pid, classname, not_classname, contains_title)
    - get_focussed_window()
    - get_ppid_of_pid(pid)
  """

  kandinsky_window_id = 0
  kandinsky_not_found_error_printed = False
  python_window_id = 0
  python_not_found_error_printed = False
  script_pid = os.getpid()
  # used for a more specific search
  script_filename = os.path.basename(sys.argv[0])
  
  # must contains this name
  winname = "kandinsky"
  # 'TkTopLevel' is the class name of root tkinter window, 'pygame' because in old releases of kandinsky i used pygame
  classnames = ("TkTopLevel", "pygame")

  def __init__(self):
    self.bind_windows()
    self.register_window_callbacks()

  def __call__(self, just_check=False):
    # check if windows still exists, to stop the KeyLogger properly
    self.check_windows_availability()
    if just_check: return False
    
    self.bind_windows()
    focussed = self.get_focussed_window()

    return ((self.python_window_id and focussed == self.python_window_id) or
            (self.kandinsky_window_id and focussed == self.kandinsky_window_id))

  def bind_windows(self):
    if (DISABLE_KANDINSKY_INPUT_ONLY or self.kandinsky_window_id == 0) and self.python_window_id == 0:
      # Find python console window and ignore the top level of tkinter
      self.python_window_id = self.get_python_console_window()

      if self.python_window_id == 0:
        # No valid (parent) window found!
        # Python probably started in no-shell-mode and/or by a task
        # So will not log python console inputs
        if not self.python_not_found_error_printed:
          prettywarn("unable to find an valid window to get inputs from python console.", RuntimeWarning)
          self.python_not_found_error_printed = True
      else: print_debug("FocusChecker", f"found the window '{self.python_window_id}' as python console")
    
    # Verify is kandinsky is imported
    if self.kandinsky_window_id == 0 and "kandinsky" in sys.modules:
      # To find kandinsky is more simple, no need to find parent processes with a valid window
      self.kandinsky_window_id = self.get_kandinsky_window()

      if self.kandinsky_window_id == 0:
        # Kandinsky window not found
        if not self.kandinsky_not_found_error_printed:
          prettywarn("could not find the kandinsky window to get inputs.", RuntimeWarning)
          self.kandinsky_not_found_error_printed = True
      else: print_debug("FocusChecker", f"found the window '{self.kandinsky_window_id}' as kandinsky")

    if self.kandinsky_window_id and not DISABLE_KANDINSKY_INPUT_ONLY:
      self.python_window_id = 0

  def check_windows_availability(self):
    if self.kandinsky_window_id == -1:
      raise RuntimeError(f"Kandinsky window destroyed. Unable to locate it.")
    if self.python_window_id == -1:
      raise RuntimeError(f"Python console window destroyed. Unable to locate it.")

  def check_window(self, wid, pid=0, classname=None, not_classname=False, contains_title=None):
    raise NotImplementedError

  def search_window(self, pid=0, classname=None, not_classname=False, contains_title=None):
    raise NotImplementedError

  def get_window(self, pid=0, classname=None, not_classname=False, contains_title=None, wid=0):
    if type(wid) != int: raise ValueError("invalid wid")
    if type(pid) != int or pid < 0: raise ValueError("invalid pid")
    if not pid and not classname and not contains_title: raise ValueError("pid, classname or contains_title must be specified")
    if classname:
      if type(classname) in (list, tuple): pass
      elif type(classname) == str: classname = (classname,)
      else: raise TypeError("invalid type for classname")
    if contains_title:
      if type(contains_title) != str: raise TypeError("invalid type for contains_title")
      contains_title = contains_title.lower()

    if wid == 0: return self.search_window(pid, classname, not_classname, contains_title)
    return wid if self.check_window(wid, pid, classname, not_classname, contains_title) else 0

  def get_ppid(self, pid):
    raise NotImplementedError

  def get_kandinsky_window(self, wid=0):
    return self.get_window(self.script_pid, self.classnames, False, self.winname, wid)

  def get_python_console_window(self, wid=0):
      wid = self.get_window(self.script_pid, self.classnames, True, self.script_filename, wid)
      if wid == 0: wid = self.get_window(self.script_pid, self.classnames, True, wid=wid)

      if wid == 0:
        # Python probably started by another process, in this mode, python don't have 'real' window
        # So try going back in the parent processes to find a valid window
        ppid = os.getppid()
        for _ in range(20): # Loop limit to avoid infinite loop
          wid = self.get_window(ppid, self.classnames, True, self.script_filename, wid)
          if wid == 0: wid = self.get_window(ppid, self.classnames, True, wid=wid)

          # Found an valid window
          if wid: break

          # Not found at this time, try with his ppid
          found_ppid = self.get_ppid(ppid)
          if found_ppid < 0: continue # error happening, will try again in the next iteration
          if found_ppid == 0: break # 0 is not a valid PID (0 is the kernel itself)
          ppid = found_ppid

      return wid

  def get_focussed_window(self):
    raise NotImplementedError

  def register_window_callbacks(self):
    return


# Fake FocusChecker class, will always return True
class NoopFocusChecker(BaseFocusChecker):
  def __init__(self): return
  def __call__(self, just_check=False): return True



from .base import *
from threading import Thread

import signal, warnings, subprocess

try:
  import Xlib
  import Xlib.xobject.drawable
except ImportError as e:
  e.msg = "Xlib module not installed. Please install it with command 'pip install python-xlib'"
  raise


# Check graphical server type
try: graphical_server_type = subprocess.check_output("loginctl show-session $(loginctl | awk '/'$(whoami)'/ {print $1}') -p Type | awk -F = '{print $2}'", shell=True, stderr=subprocess.STDOUT).decode().strip()
except subprocess.CalledProcessError as e:
  if "not been booted" in e.stdout.decode().strip(): # propably a non graphical system or no login manager
    prettywarn("no graphical server instance detected, falling back to x11 support", RuntimeWarning)
  else: prettywarn("failed to get graphical server type, falling back to x11 support", RuntimeWarning)
  # Fall baack to x11 support
  graphical_server_type = "x11"
else:
  if "not been booted" in graphical_server_type:
    prettywarn("no graphical server instance detected, falling back to x11 support", RuntimeWarning)
    graphical_server_type = "x11"

  # x11 or wayland, or... other? can be?
  if graphical_server_type not in ("x11", "wayland"):
    prettywarn(f"graphical server {graphical_server_type!r} not supported, falling back to x11 support", RuntimeWarning)

# TODO: complete support of wayland
is_wayland = graphical_server_type == "wayland"

# remove the resource warning
if ("ignore", None, ResourceWarning, None, 0) not in warnings.filters: 
  warnings.simplefilter("ignore", ResourceWarning)


class FocusChecker(BaseFocusChecker):
  # Use sys.argv[0] because the window classname of pygame if file name of script
  classnames = ("Tk", BaseFocusChecker.script_filename)

  def __init__(self):
    self.display = Xlib.display.Display()
    # Close the socket when script is finished
    signal.signal(signal.SIGINT|signal.SIGTERM|signal.SIGKILL|signal.SIGQUIT,
                  lambda: self.display.display.close_internal("client"))
    super().__init__()

  def __del__(self):
    try: self.display.close()
    except Xlib.error.ConnectionClosedError: pass # already closed
  __exit__ = __del__

  def get_wm_pid(self, window):
    p = window.get_full_property(window.display.get_atom('_NET_WM_PID'), Xlib.X.AnyPropertyType)
    if p is None: return None
    return p.value[0]

  def check_window(self, wid, pid=0, classname=None, not_classname=False, contains_title=None):
    # get the window object by his id
    if isinstance(wid, Xlib.xobject.drawable.Window): win = wid
    else: win = self.display.create_resource_object('window', wid)

    wpid = self.get_wm_pid(win)
    found = False

    if (pid == 0 or (wpid and pid == wpid)) and win.get_attributes().map_state == Xlib.X.IsViewable:
      found = True

      if found and classname:
        wclass = win.get_wm_class()
        if not wclass or not ((not_classname and any([wclass[1] != name for name in classname])) or
                          (not not_classname and any([wclass[1] == name for name in classname]))):
          found = False

      if found and contains_title:
        wtitle = win.get_wm_name()
        # check the value because some window return an empty title with this method
        if wtitle == b'':
          wtitle = win.get_full_property(self.display.get_atom('_NET_WM_NAME'), Xlib.X.AnyPropertyType)
          if wtitle: wtitle = wtitle.value.decode()
        if not wtitle or contains_title not in wtitle.lower(): found = False

    return found

  def search_window(self, pid=0, classname=None, not_classname=False, contains_title=None):
    wins = [self.display.screen().root] # should loop over all screens

    while len(wins) != 0:
      win = wins.pop(0)
      try:
        if self.check_window(win, pid, classname, not_classname, contains_title): 
          return win.id

        subwins = win.query_tree().children
        if subwins != None: wins += subwins

      except (Xlib.error.BadWindow, TypeError): pass # catch the case of a not valid window or invalid reply
      except Xlib.error.ConnectionClosedError: break # connection closed, no need to continue to search the window
    return 0

  def register_window_callbacks(self):
    def event_loop():
      # grab events
      self.display.screen().root.change_attributes(event_mask=Xlib.X.SubstructureNotifyMask)

      search_windows = True
      event = None
      
      while (self.kandinsky_window_id != -1 or
            (DISABLE_KANDINSKY_INPUT_ONLY and self.python_window_id != -1)):
        event = self.display.next_event()

        if event.type == Xlib.X.DestroyNotify:
          wid = event.window.id + 1 # idk why, i never the exact window id
          if wid == self.kandinsky_window_id: self.kandinsky_window_id = -1
          elif wid == self.python_window_id: self.python_window_id = -1

        if search_windows:
          self.bind_windows()
          if self.kandinsky_window_id and self.python_window_id: search_windows = False

    self.thread = Thread(name="WindowDestroyDetector", target=event_loop, daemon=True)
    self.thread.start()

  def get_kandinsky_window(self, wid=0):
    wid = super().get_kandinsky_window(wid)

    # In some linux distributions, Tkinter do not set window property '_NET_WM_PID'.
    # So try to find the window with a less reliable method.
    # EDIT: is in all linux distributions
    # EDIT2: Fixed in version 2.7.1 of kandinsky
    if not wid: wid = self.get_window(0, self.classnames[0], False, self.winname, wid)

    return wid

  def get_python_console_window(self, wid=0):
    # Check if is wayland because we cannot locate all windows due to "security reasons"
    if is_wayland and not self.kandinsky_not_found_error_printed:
      prettywarn("Wayland (used by GNOME, Ubuntu or KDE) is not fully supported. "
                  "The python console window will probably not be localized correctly. "
                  "To avoid this problem, start your session in X11 mode.", UserWarning)

    return super().get_python_console_window(wid)

  def get_ppid(self, pid):
    try: result = subprocess.check_output(f"ps -o ppid= {pid}".split(' ')).decode().strip()
    except subprocess.CalledProcessError: return -1

    # check the return value
    if not result: return -2
    try: return int(result)
    except ValueError: return -3

  def get_focussed_window(self):
    return self.display.screen().root.get_full_property(self.display.get_atom('_NET_ACTIVE_WINDOW'), Xlib.X.AnyPropertyType).value[0]



"""
Patch Pynput to add support of Caps Lock modifier.
"""

import sys, os

__all__ = []

pynput_backend = os.environ.get(
        'PYNPUT_BACKEND_{}'.format(__name__.rsplit('.')[-1].upper()),
        os.environ.get('PYNPUT_BACKEND', None))
if not pynput_backend: pynput_backend = sys.platform


if pynput_backend == "dummy":
  # No patch
  pass


elif pynput_backend == "uinput":
  # Idk how to test this but I hope this work
  from pynput.keyboard._uinput import Listener, Key, KeyEvent

  Listener.last_capslock_state = KeyEvent.key_up
  Listener__handle_original = Listener._handle

  def Listener__handle(self, event):
    if event.code == Key.caps_lock.value.vk:
      if event.value == KeyEvent.key_down and Listener.last_capslock_state == KeyEvent.key_up:
        if Key.shift in self._modifiers: self._modifiers.remove(Key.shift)
        else: self._modifiers.add(Key.shift)
      Listener.last_capslock_state = event.value
    Listener__handle_original(self, event)

  Listener._handle = Listener__handle


elif pynput_backend == "darwin":
  # TODO: make support
  ...


elif pynput_backend == "win32":
  # It's really poorly coded but it does the job, so I'm keeping it for now
  from pynput._util.win32 import KeyTranslator, VK, ctypes
  from pynput.keyboard._win32 import Listener

  class Capslock:
    """Class to store temporary variables for caps lock patch"""
    last_code = Listener._WM_KEYUP

    keyboard = (ctypes.c_ubyte * 255)()
    KeyTranslator._GetKeyboardState(ctypes.byref(keyboard))
    enabled = bool(keyboard[VK.CAPITAL])
    del keyboard

  Listener__convert_original = Listener._convert
  KeyTranslator__modifier_state_original = KeyTranslator._modifier_state

  def Listener__convert(self, code, msg, lpdata):
    """Wrap the hook callback to handle caps lock key code"""
    converted = Listener__convert_original(self, code, msg, lpdata)
    if converted and converted[1] == VK.CAPITAL:
      if converted[0] in Listener._PRESS_MESSAGES and converted[0] != Capslock.last_code:
        Capslock.enabled = not Capslock.enabled
      Capslock.last_code = converted[0]
    return converted
  
  def KeyTranslator__modifier_state(self):
    """Wrap the method to add the check of caps lock"""
    shift, ctrl, alt = KeyTranslator__modifier_state_original(self)
    if Capslock.enabled: shift = not shift
    return shift, ctrl, alt
  
  Listener._convert = Listener__convert
  KeyTranslator._modifier_state = KeyTranslator__modifier_state


else: # linux/xorg backend
  # For linux, this is more simple, just need to redefine one function to handle caps lock bit-mask
  import pynput.keyboard._xorg as xorg

  def shift_to_index(display, shift):
    return (          # |added thing|
        (1 if shift & 1 or shift & 3 else 0) +
        (2 if shift & xorg.alt_gr_mask(display) else 0))

  xorg.shift_to_index = shift_to_index



"""
Source: https://github.com/smokris/GetWindowID
Converted in Python by CodeConvert.ai
And modified for my own use
"""

import sys
if not sys.platform.startswith("darwin"):
  file_name = __file__[__file__.rfind("/")+1 or __file__.rfind("\\")+1:]
  raise ImportError(file_name + " can only be used on MacOS")
del sys

import Quartz


def search_window_id(requested_app, requested_window="", ignore_requested_app=False):
  session = Quartz.CGSessionCopyCurrentDictionary()
  if session: session.release()

  for window in Quartz.CGWindowListCopyWindowInfo(Quartz.kCGWindowListExcludeDesktopElements, Quartz.kCGNullWindowID):
    wid = get_window_id(window, requested_app, requested_window, ignore_requested_app)
    if wid != 0:
      return wid
  return 0

def get_window_id(window, requested_app, requested_window="", ignore_requested_app=False):
  """Return 0 if window is not the right, else his window id"""
  app = window.get("kCGWindowOwnerName", None)
  window_title = window.get("kCGWindowName", "") or window.get("kCGWindowTitle", "")
  bounds = window["kCGWindowBounds"]

  if (app != requested_app if ignore_requested_app else app == requested_app):
    aspect = bounds["Width"] / bounds["Height"]
    if aspect > 30:
      # If it's that wide and short, it's probably the system menu bar, so ignore it.
      return 0

    if requested_window:
      if not window_title:
        # If CGWindowListCopyWindowInfo didn't give us the window title, try to extract it from the Accessibility API.
        window_title = get_window_title_from_accessibility(window)

      if (window_title != requested_window or
          (window_title == "Focus Proxy" 
           and bounds["Width"] == 1 and bounds["Height"] == 1)):
        return 0
    return window["kCGWindowNumber"]
  return 0

def get_window_title_from_accessibility(window):
  ax_app = Quartz.AXUIElementCreateApplication(window["kCGWindowOwnerPID"])
  windows = Quartz.AXUIElementCopyAttributeValue(ax_app, Quartz.kAXWindowsAttribute)

  if windows == Quartz.kAXErrorSuccess:
    for ax_window in windows:
      # The Accessibility API doesn't expose a unique identifier for the window, so guess based on its frame.
      position = Quartz.AXUIElementCopyAttributeValue(ax_window, Quartz.kAXPositionAttribute)
      size = Quartz.AXUIElementCopyAttributeValue(ax_window, Quartz.kAXSizeAttribute)
      bounds = window["kCGWindowBounds"]

      if position == Quartz.kAXErrorSuccess and size == Quartz.kAXErrorSuccess:
        ax_position = Quartz.AXValueGetValue(position, Quartz.kAXValueTypeCGPoint)
        ax_size = Quartz.AXValueGetValue(size, Quartz.kAXValueTypeCGSize)

        if (ax_position.x == bounds.origin.x and ax_position.y == bounds.origin.y and
            ax_size.width == bounds.size.width and ax_size.height == bounds.size.height):
          return Quartz.AXUIElementCopyAttributeValue(ax_window, Quartz.kAXTitleAttribute)

  return ""



import warnings
warnings.filters = [] # Reset filters because some default appear in, and HIS DON'T PRINT MY WARNINGS!!

WARNINGS = False
def set_warnings(enabled):
  global WARNINGS
  WARNINGS = enabled

# prettywarn method of pysdl2
def prettywarn(message, warntype=None):
  """Prints a suppressable warning without stack or line info."""
  if not WARNINGS: return
  original = warnings.formatwarning
  warnings.formatwarning = lambda message, category, *_: f"{category.__name__}: {message}\n"
  warnings.warn(message, warntype)
  warnings.formatwarning = original

DEBUG = False
def set_debug(enabled):
  global DEBUG
  DEBUG = enabled

def print_debug(type, *msgs, **print_args):
  if DEBUG:
    if "end" in print_args and not print_args["end"].startswith('\n'): print_args["end"] += '\n'
    print(f"DEBUG: {type}: ", end='')
    print(*msgs, **print_args)


from .base import NoopFocusChecker, prettywarn, GET_INPUT_EVERYWHERE
import sys

__all__ = ["FocusChecker"]


if GET_INPUT_EVERYWHERE:
  class FocusChecker(NoopFocusChecker): ...

elif sys.platform.startswith("win"):
  from .win32 import FocusChecker
  
elif sys.platform.startswith("linux"):
  from .linux import FocusChecker
  
elif sys.platform.startswith("darwin"):
  #from .darwin import FocusChecker
  prettywarn("MacOS support is not finished, some features will be not present", ImportWarning)
  class FocusChecker(NoopFocusChecker): ...
  
else:
    # Platform not supported for focus, create an fake FocusChecker class
  # The 'focus on only window' will be disabled
  prettywarn(f"platform {sys.platform!r} not supported for focussed window inputs. "
              "Inputs will be gets on entire system", ImportWarning)
  class FocusChecker(NoopFocusChecker): ...



from .base import *
from threading import Thread

import subprocess

try:
  from AppKit import NSWorkspace
  from Quartz import CGWindowListCopyWindowInfo, kCGWindowListOptionOnScreenOnly, kCGNullWindowID
  from .darwin_GetWindowID import get_window_id
except ImportError as e:
  e.msg = ("pyobjc module and/or the Quartz extension are not installed. \n"
           "Please install them with command: pip install pyobjc-core pyobjc-framework-Quartz")
  raise


class FocusChecker(BaseFocusChecker):
  #classnames is the window owner
  classnames = ("Python",)
  #CGWindowListCreateDescriptionFromArray
  
  def check_window(self, wid, pid=0, classname=None, not_classname=False, contains_title=None):
    # get window object
    win = ...
    
    if pid == 0 or win['kCGWindowOwnerPID'] == pid:
      if classname: 
        return get_window_id(win, classname[0], not_classname, contains_title) != 0
      return True
    return False
  
  def search_window(self, pid=0, classname=None, not_classname=False, contains_title=None):
                                        # kCGWindowListExcludeDesktopElements
    for win in CGWindowListCopyWindowInfo(kCGWindowListOptionOnScreenOnly, kCGNullWindowID):
      if self.check_window(win, pid, classname, not_classname, contains_title):
        return win['kCGWindowNumber']
    return 0

  def register_window_callbacks(self):
    ...

  def get_ppid(self, pid):
    try: result = subprocess.check_output(f"ps -o ppid= {pid}".split(' ')).decode().strip()
    except subprocess.CalledProcessError: return -1

    # check the return value
    if not result: return -2
    return int(result)

  def get_focussed_window(self):
    front_app_pid = NSWorkspace.sharedWorkspace().frontmostApplication().processIdentifier()
    return self.search_window(front_app_pid) # 0 cannot happening



from setuptools import setup

def clean_pycache(path="."):
  """Clean __pycache__ directories recursively. Call this before setup()."""
  import os
  for file in os.listdir(path):
    new_path = os.path.join(path, file)
    if os.path.isdir(new_path): 
      if file == "__pycache__": 
        # Remove a file or a directory recursively using terminal commands.
        # This way avoid some permissions errors.
        if os.name == "nt": os.system(("rd /s" if os.path.isdir(new_path) else "del /f") + " /q \"" + new_path.replace('/', '\\') + "\"")
        else: os.system("rm -rf \"" + new_path.replace('\\', '/') + "\"")
      else: clean_pycache(new_path)

with open("src/ion/README.md", "rt", encoding="utf-8") as f:
  long_description = f.read()
  with open("README.md", "wt", encoding="utf-8") as f: f.write(long_description)

clean_pycache(__file__[:__file__.rfind("\\")+1 or __file__.rfind("/")+1])
setup(
  name="ion-numworks",
  version="2.0",
  author="ZetaMap",
  description="The porting of 'Ion module, from Numworks, for PC.",
  license='MIT',
  long_description=long_description,
  long_description_content_type='text/markdown',
  url="https://github.com/ZetaMap/Ion-Numworks",
  project_urls={
    "Bug Tracker": "https://github.com/ZetaMap/Ion-Numworks/issues",
  },
  classifiers=[
    'Programming Language :: Python :: 3',
    'License :: OSI Approved :: MIT License',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: Unix',
    'Operating System :: MacOS :: MacOS X', # not fully compatible
  ],
  package_dir={"": "src"},
  packages=[
    "ion",
    "ion.util",
    "ion.util.stuff",
    "ion.util.stuff.focus_checker"
  ],
  package_data={"": ["**"]},
  install_requires=["pynput"],
  python_requires= '>=3.6',
)



from os import environ
from random import randint, random

from .stuff.keylogger import *
from .stuff.keys import ALL_KEYS
from .stuff.common import *


# Enable debug
DEBUG = "ION_ENABLE_DEBUG" in environ
set_debug(DEBUG)

# Disable warnings
WARNINGS = "ION_DISABLE_WARNINGS" not in environ
set_warnings(WARNINGS)

# '0': PC, '1': Numworks, '2': Omega, '3': Upsilon
OS_MODE = environ.get('KANDINSKY_OS_MODE') or environ.get('ION_OS_MODE')
OS_MODE = (int(OS_MODE) if 0 <= int(OS_MODE) < 4 else 1) if OS_MODE and OS_MODE.isdecimal() else 1

# Check version of kandinsky to print an warning if is 'too old'
try:
  import importlib.metadata
  kandinsky_version = importlib.metadata.metadata("kandinsky").get_all("version")
  if kandinsky_version is None:
    prettywarn("invalid kandinsky metadata, are you sure the right library is installed?", DeprecationWarning)
  elif tuple([int(i) for i in kandinsky_version[0].split('.') if i.isdecimal()]) < (2, 5):
    prettywarn("for more stability, is recommended to upgrade kandinsky", DeprecationWarning)
  del importlib
except: pass


__all__ = ["Ion"]

class Ion:
  KeyLogger() # Start the key logger
  brightness = 240

  def keydown(k):
    if type(k) != int: raise TypeError(f"can't convert {type(k).__name__} to int")
    try: return KeyLogger.get_key(k)
    except IndexError: return False

  def get_keys():
    return set([k["name"] for k in ALL_KEYS if KeyLogger.get_key(k["code"])])

  # All the following functions only give a fake result to give a real look of library
  def battery(): return 4.20+randint(900, 1500)/10**5+random()/10**5
  def battery_level(): return 3
  def battery_ischarging(): return True
  def set_brightness(level):
    if type(level) != int: raise TypeError(f"can't convert {type(level).__name__} to int")
    Ion.brightness = 240 if level%256 > 240 else level%256
  def get_brightness(): return Ion.brightness

  # Caller
  def call(method, *args, **kwargs):
    try:
      print_debug("Event", method.__name__, (*args, *[f"{k}={repr(v)}" for k, v in kwargs.items()]), sep='')
      KeyLogger.raise_if_error() # raise the last KeyLogger error to main thread
      return method(*args, **kwargs), None
    except BaseException as e:
      return None, Exception.with_traceback(
        KeyboardInterrupt(type(e).__name__+": "+' '.join(e.args)) if isinstance(e, RuntimeError) else e,
        e.__traceback__.tb_next if DEBUG else None)



from .base import *
from threading import Thread

import ctypes, ctypes.wintypes, time, subprocess

class FocusChecker(BaseFocusChecker):
  def check_window(self, wid, pid=0, classname=None, not_classname=False, contains_title=None):
    if pid == 0: raise ValueError("a pid is needed")

    lpdw = ctypes.c_uint()
    ctypes.windll.user32.GetWindowThreadProcessId(wid, ctypes.byref(lpdw))

    if lpdw.value == pid and ctypes.windll.user32.IsWindowVisible(wid):
      if classname:
        buff = ctypes.create_unicode_buffer(256)
        ctypes.windll.user32.GetClassNameW(wid, buff)

        if not ((not_classname and any([buff.value != name for name in classname])) or
            (not not_classname and any([buff.value == name for name in classname]))):
          return False

      if contains_title:
        buff = ctypes.create_unicode_buffer(256)
        ctypes.windll.user32.GetWindowTextW(wid, buff, 256)

        if contains_title not in buff.value.lower(): 
          return False
      return True
    return False

  def search_window(self, pid=0, classname=None, not_classname=False, contains_title=None):
    if pid == 0: raise ValueError("a pid is needed")

    def foreach_window(hwnd, _):
      if self.check_window(hwnd, pid, classname, not_classname, contains_title):
        window.value = hwnd
        return False
      return True

    window = ctypes.c_uint(0)
    ctypes.windll.user32.EnumWindows(ctypes.WINFUNCTYPE(ctypes.c_bool, ctypes.c_uint, ctypes.c_uint)(foreach_window), 0)
    return window.value

  def register_window_callbacks(self):
    def window_state(hWinEventHook, event, hwnd, idObject, idChild, dwEventThread, dwmsEventTime):
      if idChild == idObject:
        if hwnd == self.kandinsky_window_id: self.kandinsky_window_id = -1
        elif hwnd == self.python_window_id: self.python_window_id = -1

    def register_hook():
      hook = ctypes.WINFUNCTYPE(
        ctypes.wintypes.HANDLE,
        ctypes.wintypes.HANDLE,
        ctypes.wintypes.DWORD,
        ctypes.wintypes.HWND,
        ctypes.wintypes.LONG,
        ctypes.wintypes.LONG,
        ctypes.wintypes.DWORD,
        ctypes.wintypes.DWORD
      )(window_state)
      search_windows = True
      hook_id = ctypes.windll.user32.SetWinEventHook(0x8001, 0x8001, None, hook, 0, 0, 0)

      if hook_id:
        msg = ctypes.wintypes.MSG()
        while (self.kandinsky_window_id != -1 or 
                (DISABLE_KANDINSKY_INPUT_ONLY and self.python_window_id != -1)):
          time.sleep(0.1)
          r = ctypes.windll.user32.PeekMessageW(ctypes.byref(msg), None, 0, 0, 0)

          if r == 0:
            ctypes.windll.user32.TranslateMessage(msg)
            ctypes.windll.user32.DispatchMessageW(msg)
          elif r <=0 or msg.message == 0x0401: 
            prettywarn("window destroy detector has been broken", RuntimeWarning)
            break

          if search_windows:
            self.bind_windows()
            if self.kandinsky_window_id and self.python_window_id: search_windows = False

        ctypes.windll.user32.UnhookWinEvent(hook_id)
      else: prettywarn("cannot hook the window destroy detector", RuntimeWarning)


    self.thread = Thread(name="WindowDestroyDetector", target=register_hook, daemon=True)
    self.thread.start()

  def get_ppid(self, pid):
    # Use 'wmic' command to get ppid of process
    try: result = [i.strip() for i in subprocess.check_output(f"wmic process where ProcessId={pid} get ParentProcessId".split(' '), stderr=subprocess.PIPE).decode().splitlines() if i.strip() != '']
    except subprocess.CalledProcessError: return -1

    if len(result) < 2: return -2
    return int(result[1].strip())

  def get_focussed_window(self):
    return ctypes.windll.user32.GetForegroundWindow()



KEY_LEFT: int    = 0
KEY_UP: int      = 1
KEY_DOWN: int    = 2
KEY_RIGHT: int   = 3
KEY_OK: int      = 4
KEY_BACK: int    = 5
KEY_HOME: int    = 6
KEY_ONOFF: int   = 7
KEY_SHIFT: int   = 12
KEY_ALPHA: int   = 13
KEY_XNT: int     = 14
KEY_VAR: int     = 15
KEY_TOOLBOX: int    = 16
KEY_BACKSPACE: int  = 17
KEY_EXP: int     = 18
KEY_LN: int      = 19
KEY_LOG: int     = 20
KEY_IMAGINARY: int  = 21
KEY_COMMA: int   = 22
KEY_POWER: int   = 23
KEY_SINE: int    = 24
KEY_COSINE: int  = 25
KEY_TANGENT: int = 26
KEY_PI: int      = 27
KEY_SQRT: int    = 28
KEY_SQUARE: int  = 29
KEY_SEVEN: int   = 30
KEY_EIGHT: int   = 31
KEY_NINE: int    = 32
KEY_LEFTPARENTHESIS: int  = 33
KEY_RIGHTPARENTHESIS: int = 34
KEY_FOUR: int    = 36
KEY_FIVE: int    = 37
KEY_SIX: int     = 38
KEY_MULTIPLICATION: int   = 39
KEY_DIVISION: int         = 40
KEY_ONE: int     = 42
KEY_TWO: int     = 43
KEY_THREE: int   = 44
KEY_PLUS: int    = 45
KEY_MINUS: int   = 46
KEY_ZERO: int    = 48
KEY_DOT: int     = 49
KEY_EE: int      = 50
KEY_ANS: int     = 51
KEY_EXE: int     = 52

def keydown(k: int, /) -> bool: ...
def get_keys() -> set[str]: ...
def battery() -> int: ...
def battery_level() -> int: ...
def battery_ischarging() -> int: ...
def set_brightness(level: int, /) -> None: ...
def get_brightness() -> int: ...

class file:
  SEEK_SET: int = 0
  SEEK_CUR: int = 1
  SEEK_END: int = 2
  def __init__() -> None: ...
